import re
from typing import Optional, List, Callable, Dict

import falcon

from py_fake_server.request import Request
from py_fake_server.validators import (
    WithQueryParams, WithCookies, WithBody, WithJson,
    WithContentType, WithFiles, WithHeaders, BaseValidator
)


class Statistic:
    def __init__(self, method: str, url: str):
        self.method: str = method
        self.url: str = url
        self.requests: List[Request] = []
        self._current_request_index: Optional[int] = None
        self._number_of_requests_not_specify: bool = True
        self._error_messages: List[str] = [f"Expect that server was requested with [{method.upper()}] {url}."]

    def record_request(self, request: falcon.Request):
        request_number = self.requested_times + 1
        self.requests.append(Request(request, request_number))

    @property
    def requested_times(self) -> int:
        return len(self.requests)

    def exactly_once(self) -> "Statistic":
        return self.exactly_1_times()

    def exactly_twice(self) -> "Statistic":
        return self.exactly_2_times()

    def for_the_first_time(self) -> "Statistic":
        return self.for_the_1_time()

    def for_the_second_time(self) -> "Statistic":
        return self.for_the_2_time()

    def __getattr__(self, item):
        exactly_times_pattern = r"^exactly_(?P<number>\d+)_times$"
        exactly_times_result = re.match(exactly_times_pattern, item)
        if exactly_times_result:
            number = int(exactly_times_result.groupdict()["number"])
            return self._exactly_times(number)

        for_the_time_pattern = r"^for_the_(?P<number>\d+)_time$"
        for_the_time_result = re.match(for_the_time_pattern, item)
        if for_the_time_result:
            number = int(for_the_time_result.groupdict()["number"])
            return self._for_the_time(number)

        raise AttributeError(f"'Statistic' object has no attribute '{item}'")

    def _exactly_times(self, expected_requested_times: int) -> Callable[[], "Statistic"]:
        self._number_of_requests_not_specify = False
        if expected_requested_times != self.requested_times:
            self._error_messages.append(f" {expected_requested_times} times.\n"
                                        f"But server was requested {self.requested_times} times.")
            self._raise_assertion()
        return lambda: self

    def _for_the_time(self, times: int) -> Callable[[], "Statistic"]:
        if self.requested_times < times:
            self._error_messages.append(f" At least {times} times.\n"
                                        f"But server was requested {self.requested_times} times.")
            self._raise_assertion()
        else:
            self._current_request_index = times - 1
            return lambda: self

    def with_cookies(self, cookies: Dict[str, str]) -> "Statistic":
        return self.validate(WithCookies(cookies))

    def with_body(self, body: str) -> "Statistic":
        return self.validate(WithBody(body))

    def with_json(self, json_dict: dict) -> "Statistic":
        return self.validate(WithJson(json_dict))

    def with_content_type(self, content_type: str) -> "Statistic":
        return self.validate(WithContentType(content_type))

    def with_files(self, files: Dict[str, bytes]) -> "Statistic":
        return self.validate(WithFiles(files))

    def with_headers(self, headers: Dict[str, str]) -> "Statistic":
        return self.validate(WithHeaders(headers))

    def with_query_params(self, query_params: Dict[str, str]) -> "Statistic":
        return self.validate(WithQueryParams(query_params))

    @property
    def current_request(self) -> Request:
        if self._current_request_index is None:
            raise RuntimeError("You should specify concrete request for check with 'for_the_<any_number>_time'")
        return self.requests[self._current_request_index]

    def check(self) -> bool:
        if not self.requested_times and self._number_of_requests_not_specify:
            self._error_messages.append("\nBut server was requested 0 times.")

        if self.errors_exist:
            self._raise_assertion()
        else:
            self._clean_state()
            return True

    @property
    def errors_exist(self) -> bool:
        return len(self._error_messages) > 1

    def _raise_assertion(self):
        error_message = "".join(self._error_messages)
        self._clean_state()
        raise AssertionError(error_message)

    def _clean_state(self):
        self._error_messages = self._error_messages[0:1]
        self._number_of_requests_not_specify = True
        self._current_request_index = None

    def validate(self, validator: BaseValidator) -> "Statistic":
        try:
            validator.validate(self.current_request)
        except AssertionError as error:
            self._error_messages.append(str(error))
        return self



import json as json_lib
from typing import Optional, Dict


class Response:
    def __init__(self, status: int, body: Optional[str] = None, content_type: Optional[str] = None,
                 headers: Optional[Dict[str, str]] = None, cookies: Optional[Dict[str, str]] = None,
                 json: Optional[Dict] = None):
        if status == 204 and body is not None:
            raise AttributeError("status == 204 and body != None in one response")

        headers_names = [k.lower() for k in headers.keys()] if headers else []
        if content_type and "content-type" in headers_names:
            raise AttributeError("Explicit Content-Type and Content-Type in headers in one response")
        if cookies and "cookies" in headers_names:
            raise AttributeError("Explicit Cookies and Cookies in headers in one response")
        if body is not None and json is not None:
            raise AttributeError("'body' and 'json' in one response")

        if json is not None:
            content_type = content_type or "application/json"
            body = json_lib.dumps(json)

        self.status = status
        self.body = body
        self.content_type = content_type
        self.headers = headers or {}
        self.cookies = cookies or {}



from typing import Optional, Dict

import falcon


class Request:
    def __init__(self, request: falcon.Request, request_number: int):
        self.cookies: Optional[Dict[str, str]] = request.cookies
        self.body: Optional[bytes] = request.bounded_stream.read()
        self.content_type: Optional[str] = request.content_type
        self.files: Optional[Dict[str, bytes]] = self._get_files(request)
        self.headers: Optional[Dict[str, str]] = request.headers
        self.query_params: Optional[Dict[str, str]] = request.params
        self.number = request_number

    @staticmethod
    def _get_files(request: falcon.Request) -> Optional[Dict[str, bytes]]:
        files = {
            param_name: param_value.file.read()
            for param_name, param_value in request.params.items()
            if hasattr(param_value, "file")
        }

        return files if files else None



from typing import Optional, Dict, Union

import falcon
from falcon_multipart.middleware import MultipartMiddleware
from webtest.http import StopableWSGIServer

from py_fake_server.route import Route
from py_fake_server.endpoint import Endpoint
from py_fake_server.statistic import Statistic


class FakeServer(falcon.API):
    def __init__(self, host: str, port: int):
        super().__init__(middleware=[MultipartMiddleware()])
        self.req_options = self._get_request_options()
        self._host: str = host
        self._port: int = port
        self._server: Optional[StopableWSGIServer] = None
        self._endpoints: Dict[Route, Endpoint] = {}
        self._statistics: Dict[Route, Statistic] = {}
        self.add_sink(self._handle_all)

    @staticmethod
    def _get_request_options():
        options = falcon.RequestOptions()
        options.auto_parse_qs_csv = False
        return options

    def _handle_all(self, request: falcon.Request, response: falcon.Response):
        route = Route(request.method, self.base_uri, request.path)
        endpoint = self._endpoints.get(route, Endpoint(route))

        self._set_response_attributes_from_endpoint(response, endpoint)
        self._update_statistics(request, route)

    @staticmethod
    def _set_response_attributes_from_endpoint(response: falcon.Response, endpoint: Endpoint):
        recorded_response = endpoint.pop_response()

        response.status = getattr(falcon, f"HTTP_{recorded_response.status}")
        response.body = recorded_response.body
        if recorded_response.content_type:
            response.content_type = recorded_response.content_type
        for header_name, header_value in recorded_response.headers.items():
            response.set_header(header_name, header_value)
        for cookie_name, cookie_value in recorded_response.cookies.items():
            response.set_cookie(cookie_name, cookie_value)

    def _update_statistics(self, request: falcon.Request, route: Route):
        self._statistics.setdefault(route, Statistic(route.method, route.url))
        statistic = self._statistics.get(route)
        statistic.record_request(request)

    @property
    def base_uri(self):
        return f"http://{self._host}:{self._port}"

    def start(self):
        self._server = StopableWSGIServer.create(self, host=self._host, port=self._port)

    def stop(self):
        self._server.shutdown()

    def clear(self):
        self._endpoints = {}
        self._statistics = {}

    def on_(self, method: str, url: str) -> Endpoint:
        route = Route(method, self.base_uri, url)
        new_endpoint = Endpoint(route)
        self._endpoints[route] = new_endpoint
        return new_endpoint

    def was_requested(self, method: str, url: str) -> Statistic:
        route = Route(method, self.base_uri, url)
        self._statistics.setdefault(route, Statistic(route.method, route.url))
        return self._statistics.get(route)

    def was_not_requested(self, method: str, url: str) -> Statistic:
        route = Route(method, self.base_uri, url)
        self._statistics.setdefault(route, Statistic(route.method, route.url))
        statistic: Statistic = self._statistics.get(route)
        statistic.exactly_0_times()
        return statistic


def expect_that(expectation: Union[FakeServer, Statistic]):
    if isinstance(expectation, FakeServer):
        return expectation
    else:
        return expectation.check()



import re
from typing import Optional, List, Dict

from py_fake_server.response import Response
from py_fake_server.route import Route


class Endpoint:
    def __init__(self, route: Route):
        self.method = route.method
        self.url = route.url
        self._recorded_responses: List[Response] = []
        self._error_response = Response(
            status=500,
            content_type="text/plain",
            body=f"Server has not responses for [{self.method.upper()}] {self.url}",
        )
        self._last_recorded_response_is_infinite = True

    def pop_response(self) -> Response:
        if not self._recorded_responses:
            return self._error_response

        if len(self._recorded_responses) == 1 and self._last_recorded_response_is_infinite:
            return self._recorded_responses[0]

        return self._recorded_responses.pop(0)

    def response(self, status: int, body: Optional[str] = None, content_type: Optional[str] = None,
                 headers: Optional[Dict[str, str]] = None, cookies: Optional[Dict[str, str]] = None,
                 json: Optional[Dict] = None) -> "Endpoint":

        response = Response(status, body, content_type, headers, cookies, json)
        self._recorded_responses.append(response)
        self._last_recorded_response_is_infinite = True

        return self

    def then(self) -> "Endpoint":
        return self

    def once(self) -> "Endpoint":
        return self._1_times()

    def twice(self) -> "Endpoint":
        return self._2_times()

    def __getattr__(self, item):
        times_pattern = r"^_(?P<number>\d+)_times$"
        regex_result = re.match(times_pattern, item)
        if regex_result:
            number = int(regex_result.groupdict()["number"])
            return lambda: self._times(number)
        raise AttributeError(f"'Endpoint' object has no attribute '{item}'")

    def _times(self, number: int) -> "Endpoint":
        for i in range(number - 1):
            self._recorded_responses.append(self._recorded_responses[0])

        self._last_recorded_response_is_infinite = False
        return self



class Route:
    def __init__(self, method: str, base_url: str, uri: str):
        self.method = method.lower()
        self.url = base_url + uri.rstrip("/")
        self._route_as_tuple = (self.method, self.url)

    def __hash__(self):
        return hash(self._route_as_tuple)

    def __eq__(self, other):
        return hash(self) == hash(other)



import json
from abc import ABCMeta, abstractmethod
from typing import Dict

from py_fake_server.request import Request


class HeaderDoesNotExist:
    def __repr__(self):
        return "<HEADER DOES NOT EXIST>"


class BaseValidator(metaclass=ABCMeta):
    @abstractmethod
    def validate(self, request: Request):
        pass


class WithCookies(BaseValidator):
    def __init__(self, cookies: Dict[str, str]):
        self.cookies = cookies

    def validate(self, request):
        assert self.cookies == request.cookies, \
            f"\nFor the {request.number} time: with cookies {self.cookies}.\n" \
            f"But for the {request.number} time: cookies was {request.cookies}."


class WithBody(BaseValidator):
    def __init__(self, body: str):
        self.body = body

    def validate(self, request):
        actual_body = request.body.decode("utf-8", errors="skip")
        assert self.body == actual_body, \
            f"\nFor the {request.number} time: with body {self.body.__repr__()}.\n" \
            f"But for the {request.number} time: body was {actual_body.__repr__()}."


class WithJson(BaseValidator):
    def __init__(self, json_dict: Dict):
        self.json = json_dict

    def validate(self, request):
        body = json.dumps(self.json, sort_keys=True)
        actual_body = request.body.decode("utf-8", errors="skip")

        try:
            actual_json_dict = json.loads(actual_body)
        except json.JSONDecodeError:
            assert False, \
                f"\nFor the {request.number} time: with json {body}.\n" \
                f"But for the {request.number} time: json was corrupted {actual_body.__repr__()}."

        actual_body = json.dumps(actual_json_dict, sort_keys=True)
        assert body == actual_body, \
            f"\nFor the {request.number} time: with json {body}.\n" \
            f"But for the {request.number} time: json was {actual_body}."


class WithContentType(BaseValidator):
    def __init__(self, content_type: str):
        self.content_type = content_type

    def validate(self, request):
        assert self.content_type == request.content_type, \
            f"\nFor the {request.number} time: with content type {self.content_type.__repr__()}.\n" \
            f"But for the {request.number} time: content type was {request.content_type.__repr__()}."


class WithFiles(BaseValidator):
    def __init__(self, files: Dict[str, bytes]):
        self.files = files

    def validate(self, request):
        assert self.files == request.files, \
            f"\nFor the {request.number} time: with files {self.files}.\n" \
            f"But for the {request.number} time: files was {request.files}."


class WithHeaders(BaseValidator):
    def __init__(self, headers: Dict[str, str]):
        self.headers = {name.upper(): value for name, value in headers.items()}

    def validate(self, request):
        actual_headers = request.headers
        headers_diff = self._get_headers_diff(actual_headers)

        assert not headers_diff, \
            f"\nFor the {request.number} time: with headers contain {self.headers}.\n" \
            f"But for the {request.number} time: headers contained {headers_diff}."

    def _get_headers_diff(self, actual_headers: Dict[str, str]) -> Dict[str, str]:
        headers_diff = {}

        for header_name, header_value in self.headers.items():
            if header_name in actual_headers and header_value != actual_headers[header_name]:
                headers_diff[header_name] = actual_headers[header_name]
            elif header_name not in actual_headers:
                headers_diff[header_name] = HeaderDoesNotExist()

        return headers_diff


class WithQueryParams(BaseValidator):
    def __init__(self, query_params: Dict[str, str]):
        self.query_params = query_params

    def validate(self, request):
        assert self.query_params == request.query_params, \
            f"\nFor the {request.number} time: with query params {self.query_params}.\n" \
            f"But for the {request.number} time: query params was {request.query_params}."




import codecs
import re

import os
from setuptools import (
    setup,
    find_packages,
)


def read(*parts):
    path = os.path.join(os.path.dirname(__file__), *parts)
    with codecs.open(path, encoding='utf-8') as fobj:
        return fobj.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


requires = [
    "WebTest==2.0.28",
    "falcon==1.2.0",
    "falcon-multipart==0.2.0",
]

tests_require = [
    "pytest",
    "requests",
    "pytest-cov",
]

setup(
    name="py_fake_server",
    version=find_version("py_fake_server", "__init__.py"),
    description="Make fake servers with pleasure!",
    url="https://github.com/Telichkin/py_fake_server",
    author="Roman Telichkin",
    author_email="roman@telichk.in",
    license="MIT",
    packages=find_packages(exclude=["tests"]),
    install_requires=requires,
    tests_require=tests_require,
    setup_requires=["pytest-runner"],
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
